import java.util.*;
import javax.vecmath.*;
import java.lang.Math;

public class LoopSubdivision implements SubdivisionAlgorithm
{
    /*
    private static final double ONE_EIGHTH = (1.0/8.0);
    private static final double THREE_EIGHTH = (3.0/8.0);
    private static final double SIX_EIGHTH = (6.0/8.0);
    private static final double ONE_FOURTH = (1.0/4.0);
    private static final double ONE_HALF = (1.0/2.0);
    */
    private static final double ONE_EIGHTH = 0.125;
    private static final double THREE_EIGHTH = 0.375;
    private static final double SIX_EIGHTH = 0.75;
    private static final double ONE_FOURTH = 0.25;
    private static final double ONE_HALF = 0.5;

    private Map alphaNCache = new HashMap();

    class OddVertex
    {
        int hashCode = 0;
        Point3d vertex = null;
    }

    private double getAlphaN(int valence) 
    {
        Integer key = new Integer(valence);
        Double ret = (Double)alphaNCache.get(key);
        if (ret == null) {
            ret = 
                new Double(Math.pow(THREE_EIGHTH + ONE_FOURTH * 
                            Math.cos((Math.PI * 2)/(double)valence), 2) + 
                        THREE_EIGHTH);
            alphaNCache.put(key, ret);
        }
        return ret.doubleValue();
    }
    
    private Point3d applyBoundaryMask(Mesh oldMesh,
            VertexHandle currentVertexHandle)
    {
        Point3d newPoint =
            new Point3d(currentVertexHandle.getVertex());
        newPoint.scale(SIX_EIGHTH);
        Point3d tempPoint = null;

        Iterator heIt = currentVertexHandle.oHalfEdgeIterator();
        while (heIt.hasNext()) {
            HalfEdge oHE = (HalfEdge)heIt.next();
            if (oHE.sym == null) {
                tempPoint = 
                    new Point3d(oldMesh.getVertexHandle(
                                oHE.vertexHashCode).getVertex());
                tempPoint.scale(ONE_EIGHTH);
                newPoint.add(tempPoint);
            }

            if (oHE.next.next.sym == null) {
                tempPoint = 
                    new Point3d(oldMesh.getVertexHandle(
                                oHE.next.vertexHashCode).getVertex());
                tempPoint.scale(ONE_EIGHTH);
                newPoint.add(tempPoint);
            }
        }

        return newPoint;
    }

    private Point3d applyBoundaryMaskWrong(Mesh oldMesh,
            VertexHandle currentVertexHandle)
    {
        VertexHandle tempVertexHandle;
        HalfEdge e;
        Point3d newPoint;
        Point3d tempPoint;

        //Use the boundary mask
        //(1/8)
        newPoint = new Point3d(currentVertexHandle.getVertex());
        newPoint.scale(ONE_EIGHTH);

        //e = currentHalfEdge.next;
        //We need to find which outgoing edge of this boundary vertex to
        //traverse.  Find one that will lead to the correct application 
        //of the boundary mask
        //Loop through all out-going edges, find one where next.sym
        //isn't null so we can apply the boundary mask
        Iterator it = currentVertexHandle.oHalfEdgeIterator();
        e = null;
        while (it.hasNext()) {
            e = (HalfEdge)it.next();
            if (e.next.sym != null) {
                break;
            }
            else {
                e = null;
            }
        }

        if (e == null) {
            //it is impossible to apply the boundary mask
            return newPoint;
        }

        //(3/8)
        tempVertexHandle = 
            oldMesh.getVertexHandle(e.getVertexHashCode());
        tempPoint = new Point3d(tempVertexHandle.getVertex());
        tempPoint.scale(THREE_EIGHTH);
        newPoint.add(tempPoint);

        //(3/8)
        e = e.next;
        tempVertexHandle = 
            oldMesh.getVertexHandle(e.getVertexHashCode());
        tempPoint = new Point3d(tempVertexHandle.getVertex());
        tempPoint.scale(THREE_EIGHTH);
        newPoint.add(tempPoint);

        //(1/8)
        /*
        if (e.sym == null) {
            System.out.println("WTF?????????????????????");
        }
        */
        tempPoint = 
            new Point3d(oldMesh.getVertexHandle(
                        e.sym.next.getVertexHashCode()).getVertex());
        tempPoint.scale(ONE_EIGHTH);
        newPoint.add(tempPoint);

        return newPoint;
    }

    private Point3d calcOddVertex(Mesh oldMesh,
            HalfEdge currentHalfEdge)
    {
        //First we have to find the 2 vertices the halfedge connects
        HalfEdge e = currentHalfEdge;
        while (e.next != currentHalfEdge) {
            e = e.next;
        }
        Point3d v1 = 
            new Point3d(oldMesh.getVertexHandle(
                        currentHalfEdge.getVertexHashCode()).getVertex());
        Point3d v2 = 
            new Point3d(oldMesh.getVertexHandle(
                        e.getVertexHashCode()).getVertex());

        //System.out.println("calcOddVertex() " + v1 + "-" + v2);

        Point3d newPoint;
        Point3d tempPoint;

        //If this edge is a boundary edge, then we apply the boundary
        //odd vertex rule
        if (currentHalfEdge.sym == null) {
            //(1/2)
            newPoint = new Point3d(v1);
            newPoint.scale(ONE_HALF);

            //(1/2)
            tempPoint = new Point3d(v2);
            tempPoint.scale(ONE_HALF);
            newPoint.add(tempPoint);

            //System.out.println("new odd vertex (at boundary) is " + newPoint);
            return newPoint;
        }

        /*
        System.out.println("sym points to vertex " + 
                oldMesh.getVertexHandle(currentHalfEdge.sym.getVertexHashCode()).getVertex());
                */

        //Not a boundary edge, continue normally
        //(3/8)
        newPoint = new Point3d(v1);
        //System.out.println("3/8 of " + newPoint);
        newPoint.scale(THREE_EIGHTH);
        //System.out.println("newPoint = " + newPoint);

        //(3/8)
        tempPoint = new Point3d(v2);
        //System.out.println("3/8 of " + tempPoint);
        tempPoint.scale(THREE_EIGHTH);
        newPoint.add(tempPoint);
        //System.out.println("newPoint = " + newPoint);

        //(1/8)
        tempPoint = 
            new Point3d(oldMesh.getVertexHandle(
                    currentHalfEdge.sym.next.getVertexHashCode()).getVertex());
        //System.out.println("1/8 of " + tempPoint);
        tempPoint.scale(ONE_EIGHTH);
        //System.out.println(" is " + tempPoint);
        newPoint.add(tempPoint);
        //System.out.println("newPoint = " + newPoint);

        //(1/8)
        tempPoint = 
            new Point3d(oldMesh.getVertexHandle(
                    currentHalfEdge.next.getVertexHashCode()).getVertex());
        //System.out.println("1/8 of " + tempPoint);
        tempPoint.scale(ONE_EIGHTH);
        //System.out.println(" is " + tempPoint);
        newPoint.add(tempPoint);
        //System.out.println("newPoint = " + newPoint);

        //System.out.println("new odd vertex is " + newPoint);
        return newPoint;
    }

    /**
     * Calculates an even vertex 
     */
    private Point3d calcEvenVertex(Mesh oldMesh,
            VertexHandle currentVertexHandle)
            //HalfEdge currentHalfEdge)
    {
        Point3d newPoint;
        /*
        VertexHandle currentVertexHandle = 
                    oldMesh.getVertexHandle(currentHalfEdge.
                            getVertexHashCode());
                            */

        /*
        System.out.println("calcEvenVertex() " + 
                currentVertexHandle.getVertex());
                */
        
        //check to see if this is a boundary vertex
        if (currentVertexHandle.isBoundary()) {
            //System.out.println("boundary vertex detected");
            newPoint = applyBoundaryMask(oldMesh, currentVertexHandle);
        }
        else {
            //Use the normal mask
            //Get the "one-ring" of vertices
            Iterator oneRingIt = currentVertexHandle.vertexIterator();
            int valence = 0;
            newPoint = new Point3d();

            while (oneRingIt.hasNext()) {
                valence ++;
                newPoint.add(
                        oldMesh.getVertexHandle(
                            ((Integer)oneRingIt.next()).intValue()).
                        getVertex());
            }//for all one-ring neighbours
            //calculate the new even vertex
            double alphaN =
                getAlphaN(valence);
            newPoint.scale((1.0/(double)valence) * (1.0 - alphaN));
            Point3d oldPoint = new Point3d(currentVertexHandle.getVertex());
            oldPoint.scale(alphaN);
            newPoint.add(oldPoint);
        }

        //System.out.println("new even vertex is " + newPoint);
        return newPoint;
    }

    public Mesh subdivide(Mesh oldMesh)
        throws Exception
    {
        if (!oldMesh.isTriangular()) {
            throw new Exception("LoopSubdivision requires a triangular mesh");
        }

        Mesh newMesh = new Mesh();
        Iterator oldFaceIt = oldMesh.faceIterator();
        Face oldFace;
        /*
        Point3d[] oddVertex = new Point3d[3];
        Point3d[] evenVertex = new Point3d[3];
        */
        HalfEdge currentHalfEdge;
        VertexHandle currentVertexHandle;
        int index;

        int oddVertexCounter = oldMesh.getLargestHashCode() + 1;

        while (oldFaceIt.hasNext()) {
            oldFace = (Face)oldFaceIt.next();
            //******************************************************
            /*
            System.out.println("subdividing Face : ");
            Iterator vIt = oldFace.vertexIterator();
            int numVertex = 0;
            while (vIt.hasNext()) {
                VertexHandle vH = 
                    oldMesh.getVertexHandle(vIt.next());
                System.out.println("v - " + vH.getVertex());
                numVertex++;
            }
                */
            //******************************************************
            Iterator edgeIt = oldFace.halfEdgeIterator();
            //First loop through to calculate even vertices
            index = 0;
            while (edgeIt.hasNext()) {
                //currentHalfEdge = (HalfEdge)edgeIt.next();
                currentVertexHandle = 
                    oldMesh.getVertexHandle(
                            ((HalfEdge)edgeIt.next()).getVertexHashCode());

                if (currentVertexHandle.getAttribute() == null) {
                    currentVertexHandle.setAttribute(calcEvenVertex(oldMesh,
                                currentVertexHandle));
                }

                /*
                evenVertex[index] = 
                    calcEvenVertex(oldMesh, currentHalfEdge);
                index++;
                */
            }//Done even vertices

            //Now loop through to calculate odd vertices
            edgeIt = oldFace.halfEdgeIterator();
            index = 0;
            while (edgeIt.hasNext()) {
                currentHalfEdge = (HalfEdge)edgeIt.next();
                if ((currentHalfEdge.sym == null) ||
                    (currentHalfEdge.sym.getAttribute() == null)) {
                    Point3d oddVertex =
                        calcOddVertex(oldMesh, currentHalfEdge);
                    OddVertex oddVertexAttribute = new OddVertex();
                    oddVertexAttribute.hashCode = oddVertexCounter;
                    oddVertexAttribute.vertex = oddVertex;
                    currentHalfEdge.setAttribute(oddVertexAttribute);
                    //System.out.println("oddVertexCounter " + oddVertexCounter);
                    oddVertexCounter ++;
                }
                else {
                    currentHalfEdge.setAttribute(
                            currentHalfEdge.sym.getAttribute());
                }
                /*
                oddVertex[index] = 
                    calcOddVertex(oldMesh, currentHalfEdge);
                index++;
                */
            }//Done odd vertices
            //Add the four new triangles in the new mesh
            /*
            VertexHandle []evenHandle = new VertexHandle[3];
            evenHandle[0] = newMesh.addVertex(evenVertex[0]);
            evenHandle[1] = newMesh.addVertex(evenVertex[1]);
            evenHandle[2] = newMesh.addVertex(evenVertex[2]);
            
            VertexHandle []oddHandle = new VertexHandle[3];
            oddHandle[0] = newMesh.addVertex(oddVertex[0]);
            oddHandle[1] = newMesh.addVertex(oddVertex[1]);
            oddHandle[2] = newMesh.addVertex(oddVertex[2]);

            VertexHandle []newFace = new VertexHandle[3];

            newFace[0] = oddHandle[0];
            newFace[1] = oddHandle[2];
            newFace[2] = evenHandle[2];
            newMesh.addFace(newFace);

            newFace[0] = oddHandle[1];
            newFace[1] = oddHandle[0];
            newFace[2] = evenHandle[0];
            newMesh.addFace(newFace);
            
            newFace[0] = oddHandle[1];
            newFace[1] = oddHandle[2];
            newFace[2] = oddHandle[0];
            newMesh.addFace(newFace);

            newFace[0] = oddHandle[2];
            newFace[1] = oddHandle[1];
            newFace[2] = evenHandle[1];
            newMesh.addFace(newFace);
            */
            //TESTING____________________________________________
            //break;
        }//for all old faces


        //Now cycle through the old mesh again, "spawning" a new mesh
        oldFaceIt = oldMesh.faceIterator();
        while (oldFaceIt.hasNext()) {
            oldFace = (Face)oldFaceIt.next();
            Iterator edgeIt = oldFace.halfEdgeIterator();
            Point3d[] newVertices = new Point3d[6];
            VertexHandle[] newHandles = new VertexHandle[6];
            VertexHandle[] newFaces = new VertexHandle[3];
            index = 0;
            while (edgeIt.hasNext()) {
                HalfEdge he = (HalfEdge)edgeIt.next();

                OddVertex oddVertexAttribute =
                    (OddVertex)he.getAttribute();
                //newVertices[index] = oddVertexAttribute.vertex;
                newHandles[index] = 
                    newMesh.addVertex(new Point3d(oddVertexAttribute.vertex),
                            oddVertexAttribute.hashCode);
                /*
                newVertices[index] = (Point3d)he.getAttribute();
                newHandles[index] = newMesh.addVertex(newVertices[index]);
                */
                index++;
                newVertices[index] = 
                    (Point3d)(oldMesh.getVertexHandle(
                            he.getVertexHashCode()).getAttribute());
                newHandles[index] = 
                    newMesh.addVertex(new Point3d(newVertices[index]), 
                            he.getVertexHashCode());
                index++;
            }
            //now insert the 4 triangles into the new mesh
            newFaces[0] = newHandles[4];
            newFaces[1] = newHandles[2];
            newFaces[2] = newHandles[3];
            newMesh.addFace(newFaces);

            newFaces[0] = newHandles[5];
            newFaces[1] = newHandles[0];
            newFaces[2] = newHandles[4];
            newMesh.addFace(newFaces);

            newFaces[0] = newHandles[0];
            newFaces[1] = newHandles[2];
            newFaces[2] = newHandles[4];
            newMesh.addFace(newFaces);

            newFaces[0] = newHandles[0];
            newFaces[1] = newHandles[1];
            newFaces[2] = newHandles[2];
            newMesh.addFace(newFaces);
            /*
            newFaces[0] = newHandles[4];
            newFaces[1] = newHandles[3];
            newFaces[2] = newHandles[2];
            newMesh.addFace(newFaces);

            newFaces[0] = newHandles[5];
            newFaces[1] = newHandles[4];
            newFaces[2] = newHandles[0];
            newMesh.addFace(newFaces);

            newFaces[0] = newHandles[4];
            newFaces[1] = newHandles[2];
            newFaces[2] = newHandles[0];
            newMesh.addFace(newFaces);

            newFaces[0] = newHandles[0];
            newFaces[1] = newHandles[2];
            newFaces[2] = newHandles[1];
            newMesh.addFace(newFaces);
            */
        }

        return newMesh;
    }
}
