import java.util.*;
import javax.vecmath.*;
import java.lang.Math;

public class ModifiedButterflySubdivision implements SubdivisionAlgorithm
{
    /*
    private static final double THREE_EIGHTH = 0.375;
    private static final double SIX_EIGHTH = 0.75;
    */

    class OddVertex
    {
        int hashCode = 0;
        Point3d vertex = null;
    }

    private int oddVertexCounter = 0;

    private static final double ONE_EIGHTH = 0.125;
    private static final double ONE_FOURTH = 0.25;
    private static final double ONE_HALF = 0.5;
    private static final double THREE_EIGHTH = 0.375;
    private static final double NEGATIVE_ONE_EIGHTH = (-0.125);
    private static final double NEGATIVE_ONE_SIXTEENTH = (-0.0625);
    private static final double NINE_SIXTEENTH = 0.5625;
    private static final double FIVE_TWELVETH = (5.0/12.0);
    private static final double NEGAVIVE_ONE_TWELVETH = (-1.0/12.0);
    private static final double THREE_FOURTH = 0.75;

    private static double[] TEN_POINT_MASK;


    private static Map sJCache = new HashMap();

    public ModifiedButterflySubdivision()
    {
        TEN_POINT_MASK = new double[6];
        TEN_POINT_MASK[0] = ONE_HALF;
        TEN_POINT_MASK[1] = ONE_EIGHTH;
        TEN_POINT_MASK[2] = NEGATIVE_ONE_SIXTEENTH;
        TEN_POINT_MASK[3] = 0.;
        TEN_POINT_MASK[4] = NEGATIVE_ONE_SIXTEENTH;
        TEN_POINT_MASK[5] = ONE_EIGHTH;
    }

    private double getSj(int j, int valence) 
    {
        Integer valenceKey = new Integer(valence);
        Integer jKey = new Integer(j);
        Map jMap = (Map)sJCache.get(valenceKey);
        Double ret = null;
        if (jMap != null) {
            ret = (Double)jMap.get(jKey);
        }
        if (ret == null) {
            ret = 
                new Double( ((ONE_FOURTH) + 
                            Math.cos((Math.PI * 2. * (double)j)/
                                (double)valence) + 
                            ((ONE_HALF) * 
                             Math.cos((Math.PI * 4. * (double)j)/
                                 (double)valence)))/
                        (double)valence);
            if (jMap == null) {
                jMap = new HashMap();
            }
            jMap.put(jKey, ret);
            sJCache.put(valenceKey, jMap);
        }
        return ret.doubleValue();
    }

    private Point3d applyExtrodinaryMask(Mesh oldMesh, 
            HalfEdge currentHalfEdge, int valence)
    {
        Point3d newPoint = null;
        Point3d tempPoint = null;
        VertexHandle q =
            oldMesh.getVertexHandle(
                    currentHalfEdge.next.next.getVertexHashCode());

        if (valence == 3) {
            Iterator vIt = q.vertexIterator(currentHalfEdge);
            int index = 0;
            newPoint = new Point3d();
            while (vIt.hasNext()) {
                tempPoint = 
                    new Point3d(
                            oldMesh.getVertexHandle(((Integer)vIt.next()).intValue()).getVertex());
                if (index == 0) {
                    tempPoint.scale(FIVE_TWELVETH);
                    newPoint.add(tempPoint);
                }
                else {
                    tempPoint.scale(NEGAVIVE_ONE_TWELVETH);
                    newPoint.add(tempPoint);
                }
                index ++;
            }
        }
        else if (valence == 4) {
            Iterator vIt = q.vertexIterator(currentHalfEdge);
            int index = 0;
            newPoint = new Point3d();
            while (vIt.hasNext()) {
                tempPoint = 
                    new Point3d(
                            oldMesh.getVertexHandle(((Integer)vIt.next()).intValue()).getVertex());
                if (index == 0) {
                    //System.out.println("at index 0 is " + tempPoint);
                    tempPoint.scale(THREE_EIGHTH);
                    newPoint.add(tempPoint);
                }
                else if (index == 2) {
                    //System.out.println("at index 2 is " + tempPoint);
                    tempPoint.scale(NEGATIVE_ONE_EIGHTH);
                    newPoint.add(tempPoint);
                }
                index ++;
            }
        }
        else if (valence >= 5) {
            Iterator vIt = q.vertexIterator(currentHalfEdge);
            int index = 0;
            newPoint = new Point3d();
            while (vIt.hasNext()) {
                tempPoint = 
                    new Point3d(
                            oldMesh.getVertexHandle(((Integer)vIt.next()).intValue()).getVertex());
                tempPoint.scale(getSj(index, valence));
                newPoint.add(tempPoint);
                index ++;
            }
        }
        else {
            //wtf????
            //throw new RuntimeException();
            newPoint = new Point3d(q.getVertex());
            newPoint.scale(ONE_FOURTH);
        }

        Point3d qPoint = new Point3d(q.getVertex());
        qPoint.scale(THREE_FOURTH);
        newPoint.add(qPoint);
        return newPoint;
    }
    
    private Point3d calcOddVertex(Mesh oldMesh,
            HalfEdge currentHalfEdge)
    {
        //First we have to find the 2 vertices the halfedge connects
        HalfEdge e = currentHalfEdge;
        while (e.next != currentHalfEdge) {
            e = e.next;
        }
        VertexHandle prevVertexHandle =
            oldMesh.getVertexHandle(e.getVertexHashCode());
        VertexHandle nextVertexHandle =
            oldMesh.getVertexHandle(currentHalfEdge.getVertexHashCode());

        Point3d prevPoint = new Point3d(prevVertexHandle.getVertex());
        Point3d nextPoint = new Point3d(nextVertexHandle.getVertex());

        Point3d newPoint;
        Point3d tempPoint;

        //If this edge is a boundary edge, then we apply the boundary
        //odd vertex rule
        if (currentHalfEdge.sym == null) {
            //System.out.println("applying boundary edge rule");
            //We need four points for the boundary mask
            Point3d prevPrevPoint = null;
            Point3d nextNextPoint = null;

            Iterator heIt = prevVertexHandle.oHalfEdgeIterator();
            //Set prevPrev
            while (heIt.hasNext()) {
                HalfEdge oHE = (HalfEdge)heIt.next();
                if (oHE.next.next.sym == null) {
                    prevPrevPoint = 
                        new Point3d(oldMesh.getVertexHandle(
                                    oHE.next.vertexHashCode).getVertex());
                    break;
                }
            }
            heIt = nextVertexHandle.oHalfEdgeIterator();
            //Set nextNext point
            while (heIt.hasNext()) {
                HalfEdge oHE = (HalfEdge)heIt.next();
                if (oHE.sym == null) {
                    nextNextPoint = 
                        new Point3d(oldMesh.getVertexHandle(
                                    oHE.vertexHashCode).getVertex());
                    break;
                }
            }

            /*
            System.out.println("prevPoint - " + prevPoint);
            System.out.println("nextPoint - " + nextPoint);
            System.out.println("prevPrevPoint - " + prevPrevPoint);
            System.out.println("nextNextPoint - " + nextNextPoint);
            */

            //(-1/16)
            newPoint = new Point3d(prevPrevPoint);
            newPoint.scale(NEGATIVE_ONE_SIXTEENTH);

            //(9/16)
            tempPoint = new Point3d(prevPoint);
            tempPoint.scale(NINE_SIXTEENTH);
            newPoint.add(tempPoint);

            //(9/16)
            tempPoint = new Point3d(nextPoint);
            tempPoint.scale(NINE_SIXTEENTH);
            newPoint.add(tempPoint);

            //(-1/16)
            tempPoint = new Point3d(nextNextPoint);
            tempPoint.scale(NEGATIVE_ONE_SIXTEENTH);
            newPoint.add(tempPoint);

            //System.out.println("new edge point is " + newPoint);
            return newPoint;
        }

        int prevValence = prevVertexHandle.getValence();
        int nextValence = nextVertexHandle.getValence();

        /*
        System.out.println("prevPoint = " + prevPoint + 
                ", nextPoint = " + nextPoint);
        System.out.println("prevValance is " + prevValence +
                ", nextValence is " + nextValence);
                */

        if (prevValence == 6 && nextValence == 6) {
            //both vertices are of valence 6
            //
            //Do the left butterfly first
            Iterator vIt = prevVertexHandle.vertexIterator(currentHalfEdge);
            int index = 0;
            newPoint = new Point3d();
            while (vIt.hasNext()) {
                tempPoint = 
                    new Point3d(
                            oldMesh.getVertexHandle(((Integer)vIt.next()).intValue()).getVertex());
                /*
                System.out.println("index is " + index + " point is " + 
                        tempPoint);
                        */
                tempPoint.scale(TEN_POINT_MASK[index]);
                /*
                System.out.println("after scale is " + tempPoint);
                System.out.println("---------------------------------");
                */
                newPoint.add(tempPoint);
                index ++;
            }

            //now do the right butterfly
            vIt = nextVertexHandle.vertexIterator(currentHalfEdge.sym);
            index = 0;
            while (vIt.hasNext()) {
                tempPoint = 
                    new Point3d(
                            oldMesh.getVertexHandle(((Integer)vIt.next()).intValue()).getVertex());
                /*
                System.out.println("index is " + index + " point is " + 
                        tempPoint);
                        */
                if (index == 1 || index == 5) {
                    index ++;
                    continue;
                }
                tempPoint.scale(TEN_POINT_MASK[index]);
                /*
                System.out.println("after scale is " + tempPoint);
                System.out.println("---------------------------------");
                */
                newPoint.add(tempPoint);
                index ++;
            }
        }
        else {
            //Either or both vertices is extrodinary
            Point3d p1 = null;
            Point3d p2 = null;
            if (prevValence != 6) {
                p1 = 
                    applyExtrodinaryMask(oldMesh, currentHalfEdge, prevValence);
            }
            if (nextValence != 6) {
                p2 = 
                    applyExtrodinaryMask(oldMesh, currentHalfEdge.sym, 
                            nextValence);
            }
            
            if (p1 != null) {
                if (p2 != null) {
                    p1.add(p2);
                    p1.scale(ONE_HALF);
                }
                newPoint = p1;
            }
            else {
                newPoint = p2;
            }
        }
        //System.out.println("new edge point is " + newPoint);
        return newPoint;
    }

    public Mesh subdivide(Mesh oldMesh)
        throws Exception
    {
        if (!oldMesh.isTriangular()) {
            throw new Exception("ModifiedButterflySubdivision requires a triangular mesh");
        }

        oddVertexCounter = oldMesh.getLargestHashCode() + 1;
        Mesh newMesh = new Mesh();
        Iterator oldFaceIt = oldMesh.faceIterator();
        Face oldFace;
        Point3d[] oddVertex = new Point3d[3];
        Point3d[] evenVertex = new Point3d[3];
        HalfEdge currentHalfEdge;
        VertexHandle currentVertexHandle;
        int index;

        while (oldFaceIt.hasNext()) {
            oldFace = (Face)oldFaceIt.next();
            Iterator edgeIt = oldFace.halfEdgeIterator();
            //Now loop through to calculate odd vertices
            edgeIt = oldFace.halfEdgeIterator();
            index = 0;
            while (edgeIt.hasNext()) {
                currentHalfEdge = (HalfEdge)edgeIt.next();
                if ((currentHalfEdge.sym == null) ||
                    (currentHalfEdge.sym.getAttribute() == null)) {
                    OddVertex ov = new OddVertex();
                    ov.vertex = calcOddVertex(oldMesh, currentHalfEdge);
                    ov.hashCode = oddVertexCounter;
                    oddVertexCounter++;
                    currentHalfEdge.setAttribute(ov);
                    /*
                    currentHalfEdge.setAttribute(calcOddVertex(oldMesh,
                                currentHalfEdge));
                                */
                }
                else {
                    currentHalfEdge.setAttribute(
                            currentHalfEdge.sym.getAttribute());
                }
            }//Done odd vertices
        }//for all old faces


        //Now cycle through the old mesh again, "spawning" a new mesh
        oldFaceIt = oldMesh.faceIterator();
        while (oldFaceIt.hasNext()) {
            oldFace = (Face)oldFaceIt.next();
            Iterator edgeIt = oldFace.halfEdgeIterator();
            Point3d[] newVertices = new Point3d[6];
            VertexHandle[] newHandles = new VertexHandle[6];
            VertexHandle[] newFaces = new VertexHandle[3];
            index = 0;
            while (edgeIt.hasNext()) {
                Point3d newVertex;
                HalfEdge he = (HalfEdge)edgeIt.next();
                OddVertex ov = (OddVertex)he.getAttribute();
                newVertex = new Point3d(ov.vertex);
                newHandles[index] = newMesh.addVertex(newVertex, ov.hashCode);
                /*
                newVertex = new Point3d((Point3d)he.getAttribute());
                newHandles[index] = newMesh.addVertex(newVertex);
                */
                index++;

                newVertex = 
                    new Point3d(oldMesh.getVertexHandle(he.getVertexHashCode()).getVertex());
                newHandles[index] = 
                    newMesh.addVertex(newVertex, he.getVertexHashCode());
                index++;
            }
            //now insert the 4 triangles into the new mesh
            newFaces[0] = newHandles[4];
            newFaces[1] = newHandles[2];
            newFaces[2] = newHandles[3];
            newMesh.addFace(newFaces);

            newFaces[0] = newHandles[5];
            newFaces[1] = newHandles[0];
            newFaces[2] = newHandles[4];
            newMesh.addFace(newFaces);

            newFaces[0] = newHandles[0];
            newFaces[1] = newHandles[2];
            newFaces[2] = newHandles[4];
            newMesh.addFace(newFaces);

            newFaces[0] = newHandles[0];
            newFaces[1] = newHandles[1];
            newFaces[2] = newHandles[2];
            newMesh.addFace(newFaces);
        }

        return newMesh;
    }
}
