import java.util.*;
import javax.vecmath.*;
import java.lang.Math;

public class PetersSubdivision implements SubdivisionAlgorithm
{
    public class HalfEdgeAttributes
    {
        Point3d midPoint = null;
        int hashCode = 0;
    }

    public class VertexAttributes
    {
        boolean createdFace = true;
    }

    private static final double ONE_HALF = 0.5;
    private int oddVertexCounter = 0;


    private Point3d calcMidPoint(Mesh oldMesh, HalfEdge currentHalfEdge)
    {
        Point3d prevPoint;
        Point3d nextPoint;
        HalfEdge prevEdge = currentHalfEdge;

        nextPoint = 
            new Point3d(oldMesh.getVertexHandle(
                    currentHalfEdge.getVertexHashCode()).getVertex());

        while (prevEdge.next != currentHalfEdge) {
            prevEdge = prevEdge.next;
        }
        prevPoint = 
            new Point3d(oldMesh.getVertexHandle(
                    prevEdge.getVertexHashCode()).getVertex());

        prevPoint.add(nextPoint);
        prevPoint.scale(ONE_HALF);

        return prevPoint;
    }

    public Mesh subdivide(Mesh oldMesh)
        throws Exception
    {
        Mesh newMesh = new Mesh();
        Iterator oldFaceIt = oldMesh.faceIterator();
        Face oldFace;
        oddVertexCounter = oldMesh.getLargestHashCode() + 1;

        while (oldFaceIt.hasNext()) {
            oldFace = (Face)oldFaceIt.next();
            Iterator edgeIt = oldFace.halfEdgeIterator();
            //loop through to calculate odd vertices
            while (edgeIt.hasNext()) {
                HalfEdgeAttributes hea = null;
                HalfEdge currentHalfEdge = (HalfEdge)edgeIt.next();
                if (currentHalfEdge.sym != null) {
                    hea = 
                        (HalfEdgeAttributes)(currentHalfEdge.sym.getAttribute());
                }
                if (hea == null) {
                    Point3d midPoint =
                        calcMidPoint(oldMesh, currentHalfEdge);
                    hea = new HalfEdgeAttributes();
                    hea.midPoint = midPoint;
                    hea.hashCode = oddVertexCounter;
                    oddVertexCounter ++;
                }
                currentHalfEdge.setAttribute(hea);
            }
        }//for all old faces


        //Now cycle through the old mesh again, "spawning" a new mesh
        oldFaceIt = oldMesh.faceIterator();
        while (oldFaceIt.hasNext()) {
            oldFace = (Face)oldFaceIt.next();
            Iterator edgeIt = oldFace.halfEdgeIterator();
            List faceFaceVertex = new LinkedList();

            while (edgeIt.hasNext()) {
                HalfEdge he = (HalfEdge)edgeIt.next();
                VertexHandle oldVertex = 
                    oldMesh.getVertexHandle(he.getVertexHashCode());

                faceFaceVertex.add(
                        ((HalfEdgeAttributes)he.getAttribute()));

                if (oldVertex.getAttribute() == null) {
                    List vertexFaceVertex = new LinkedList();
                    //go ahead and spin around this old vertex to create
                    //a face
                    Iterator oHeIt = oldVertex.oHalfEdgeIterator();
                    while (oHeIt.hasNext()) {
                        HalfEdgeAttributes hea =
                            (HalfEdgeAttributes)
                            ((HalfEdge)oHeIt.next()).getAttribute();

                        vertexFaceVertex.add(hea);
                    }

                    //Add the vertices into the new mesh
                    VertexHandle[] vh = 
                        new VertexHandle[vertexFaceVertex.size()];
                    for (int a = 0; a < vertexFaceVertex.size(); a++) {
                        HalfEdgeAttributes hea = 
                            (HalfEdgeAttributes)vertexFaceVertex.get(a);
                        vh[a] = 
                            newMesh.addVertex(hea.midPoint, hea.hashCode);
                    }
                    //add the face for a vertex
                    newMesh.addFace(vh);
                    oldVertex.setAttribute(new VertexAttributes());
                }//if we havn't created a face for this even vertex yet
            }//for all edges in this oldFace

            //now add a face for the face
            if (faceFaceVertex.size() < 3) {
                continue;
            }
            VertexHandle[] vh = 
                new VertexHandle[faceFaceVertex.size()];
            //System.out.println("face face has size " + faceFaceVertex.size());
            for (int a = 0; a < faceFaceVertex.size(); a++) {
                HalfEdgeAttributes hea = 
                    (HalfEdgeAttributes)faceFaceVertex.get(a);
                vh[a] = 
                    newMesh.addVertex(hea.midPoint, hea.hashCode);
                /*
                vh[a] = 
                    newMesh.addVertex((Point3d)faceFaceVertex.get(a));
                    */
            }
            newMesh.addFace(vh);
        }//for all faces in the old mesh

        return newMesh;
    }
}
