import java.io.*;
import java.util.*;
import javax.vecmath.*;

import javax.swing.*;
import javax.swing.border.*;
import javax.swing.event.*;

import java.awt.*;
import java.awt.event.*;

public class SubdivisionController
{
    public static final String APP_TITLE = "Subdivision Controller";
    private JFrame topLevel;

    private File selectedFile = null;
    private SubdivisionAlgorithm selectedAlgorithm = null;
    private int subdivisionDepth = 0;
    private Mesh mesh = null;
    private Class algorithmClass;

    private JTextField numPolygonsTL;
    private JTextField subdivisionDepthTL;
    private JTextField algorithmSelectedTL;
    private JTextField fileOpenedTL;
    
    private JCheckBox triangulate;


    public void loadFromIV(File f)
        throws Exception
    {
        mesh = new Mesh(triangulate.isSelected());

        //state of 0 means we're probing
        //state of 1 means we're reading vertices
        //state of 2 means we're reading faces
        int state = 0;
        int vhIndex = 0;
        java.util.List vh = new ArrayList(500);

        BufferedReader br = new BufferedReader(new FileReader(f));
        while (br.ready()) {
            String currentLine = br.readLine();
            //StringTokenizer stk = new StringTokenizer(currentLine);
            if (currentLine.indexOf("]") != -1) {
                state = 0;
                continue;
            }

            if (currentLine.length() == 0) {
                continue;
            }
            //String keyWord = stk.nextToken();


            if (state == 0) {
                if (currentLine.indexOf("point") != -1) {
                    state = 1;
                    continue;
                }
                else if (currentLine.indexOf("coordIndex") != -1) {
                    state = 2;
                    continue;
                }
                else {
                    continue;
                }
            }

            switch(state)
            {
                case 1:
                    {
                        StringTokenizer st = new StringTokenizer(currentLine);
                        //skip the keyWord 'v'
                        //st.nextToken();
                        String x = st.nextToken();
                        String y = st.nextToken();
                        String z = st.nextToken();
                        int comma = z.indexOf(",");
                        if (comma != -1) {
                            z = z.substring(0, comma);
                        }

                        Point3d v = 
                            new Point3d(Double.parseDouble(x),
                                    Double.parseDouble(y),
                                    Double.parseDouble(z));

                        vh.add(mesh.addVertex(v, vhIndex));
                        vhIndex++;
                        //System.out.println(""+ v);
                        break;
                    }
                case 2:
                    {
                        StringTokenizer st = 
                            new StringTokenizer(currentLine, ", \t\n\r\f");
                        //skip the keyWord 'f'
                        //st.nextToken();
                        //System.out.println("face  " + st.countTokens());
                        VertexHandle []face = 
                            new VertexHandle[st.countTokens() - 1];
                        for (int a = 0; a < face.length; a++) {
                            String indexStr = st.nextToken();
                            //System.out.println("index is " + indexStr);
                            int index = 
                                Integer.parseInt(indexStr);
                            //since our array starts at 0, but their array 
                            //starts at 1
                            face[a] = 
                                (VertexHandle)vh.get(index);
                        }
                        try {
                            mesh.addFace(face);
                        }
                        catch (Exception e) {
                            System.out.println("fucked up trying to add " + currentLine);
                            throw e;
                        }
                        break;
                    }
            }//switch(state)
        }//while the file hasn't been all parsed yet
        br.close();
    }

    public void loadFromM(File f)
        throws Exception
    {
        mesh = new Mesh(triangulate.isSelected());

        //state of 0 means we're reading vertices
        //state of 1 means we're reading faces
        int state = 0;
        Map vh = new HashMap();

        BufferedReader br = new BufferedReader(new FileReader(f));
        while (br.ready()) {
            String currentLine = br.readLine();
            StringTokenizer stk = new StringTokenizer(currentLine);

            if (currentLine.length() == 0) {
                continue;
            }
            String keyWord = stk.nextToken();

            if (keyWord.equals("Vertex")) {
                state = 0;
            }
            else if (keyWord.equals("Face")) {
                state = 1;
            }
            else {
                continue;
            }

            switch(state)
            {
                case 0:
                    {
                        StringTokenizer st = new StringTokenizer(currentLine);
                        //skip the keyWord 'Vertex'
                        st.nextToken();
                        String id = st.nextToken();
                        Point3d v = 
                            new Point3d(Double.parseDouble(st.nextToken()),
                                    Double.parseDouble(st.nextToken()),
                                    Double.parseDouble(st.nextToken()));

                        vh.put(new Integer(id), 
                                mesh.addVertex(v, Integer.parseInt(id)));

                        break;
                    }
                case 1:
                    {
                        StringTokenizer st = new StringTokenizer(currentLine);
                        //skip the keyWord 'Face'
                        st.nextToken();
                        //skip the face ID
                        st.nextToken();
                        VertexHandle []face = 
                            new VertexHandle[st.countTokens()];
                        for (int a = 0; a < face.length; a++) {
                            int index = 
                                Integer.parseInt(st.nextToken());
                            face[a] = 
                                (VertexHandle)vh.get(new Integer(index));
                        }
                        try {
                            mesh.addFace(face);
                        }
                        catch (Exception e) {
                            System.out.println("fucked up trying to add " + currentLine);
                            throw e;
                        }
                        break;
                    }
            }//switch(state)
        }//while the file hasn't been all parsed yet
        br.close();
    }

    private void loadFromObj(File f)
        throws Exception
    {
        mesh = new Mesh(triangulate.isSelected());

        //state of 0 means we're reading vertices
        //state of 1 means we're reading faces
        int state = 0;
        int vhIndex = 1;
        java.util.List vh = new ArrayList(500);

        BufferedReader br = new BufferedReader(new FileReader(f));
        while (br.ready()) {
            String currentLine = br.readLine();
            StringTokenizer stk = new StringTokenizer(currentLine);

            if (currentLine.length() == 0) {
                continue;
            }
            String keyWord = stk.nextToken();

            if (keyWord.equals("v")) {
                state = 0;
            }
            else if (keyWord.equals("f")) {
                state = 1;
            }
            else {
                continue;
            }

            switch(state)
            {
                case 0:
                    {
                        StringTokenizer st = new StringTokenizer(currentLine);
                        //skip the keyWord 'v'
                        st.nextToken();
                        Point3d v = 
                            new Point3d(Double.parseDouble(st.nextToken()),
                                    Double.parseDouble(st.nextToken()),
                                    Double.parseDouble(st.nextToken()));

                        vh.add(mesh.addVertex(v, vhIndex));
                        vhIndex++;
                        break;
                    }
                case 1:
                    {
                        StringTokenizer st = new StringTokenizer(currentLine);
                        //skip the keyWord 'f'
                        st.nextToken();
                        VertexHandle []face = 
                            new VertexHandle[st.countTokens()];
                        for (int a = 0; a < face.length; a++) {
                            String vertexIndex = st.nextToken();
                            StringTokenizer vertexIndexSt =
                                new StringTokenizer(vertexIndex, "/");
                            int index = 
                                Integer.parseInt(vertexIndexSt.nextToken());
                            //since our array starts at 0, but their array 
                            //starts at 1
                            face[a] = 
                                (VertexHandle)vh.get(index - 1);
                        }
                        try {
                        mesh.addFace(face);
                        }
                        catch (Exception e) {
                            System.out.println("fucked up trying to add " + currentLine);
                            throw e;
                        }
                        break;
                    }
            }//switch(state)
        }//while the file hasn't been all parsed yet
        br.close();
    }

    private void loadFromS3D(File f)
        throws Exception
    {
        mesh = new Mesh(triangulate.isSelected());

        BufferedReader br = new BufferedReader(new FileReader(f));
        int vertexCounter = 0;
        Map vertexToVertexIndexMap = new HashMap();
        while (br.ready()) {
            String patchDefinition = br.readLine();

            if (patchDefinition.trim().equals("") ||
                    patchDefinition.startsWith("#")) {
                continue;
            }
            StringTokenizer st = new StringTokenizer(patchDefinition);
            //We are only interested in the number of vertices
            st.nextToken();
            int numVertex = Integer.parseInt(st.nextToken());
            VertexHandle[] vertexHandle = new VertexHandle[numVertex];

            for (int a = 0; a < numVertex; a++) {
                //Skip the normal
                String vertex = br.readLine();
                st = new StringTokenizer(vertex);
                String lineType = st.nextToken();

                if (lineType.equals("n")) {
                    vertex = br.readLine();
                    st = new StringTokenizer(vertex);
                    //Skip the "v" prefix
                    st.nextToken();
                }

                Point3d v = 
                    new Point3d(Double.parseDouble(st.nextToken()),
                            Double.parseDouble(st.nextToken()),
                            Double.parseDouble(st.nextToken()));

                Integer vertexIndex = 
                    (Integer)vertexToVertexIndexMap.get(v.toString());
                if (vertexIndex == null) {
                    vertexToVertexIndexMap.put(v.toString(), 
                            new Integer(vertexCounter));
                    vertexIndex = new Integer(vertexCounter);
                    vertexCounter ++;
                }

                vertexHandle[a] = mesh.addVertex(v, vertexIndex.intValue());
                if (!vertexHandle[a].getVertex().equals(v)) {
                    throw new RuntimeException();
                }
            }
            mesh.addFace(vertexHandle);
            //skip the "ending" line
            br.readLine();
        }
        br.close();
    }
    private void subdivide()
    {
        if (selectedAlgorithm == null) {
            JOptionPane.showMessageDialog(topLevel, 
                    "Select an algorithm first!");
            return;
        }
        try {
            mesh = selectedAlgorithm.subdivide(mesh);
            subdivisionDepth ++;
        }
        catch (Exception e) {
            JOptionPane.showMessageDialog(topLevel, 
                    e.toString());
        }
    }

    private void openFile()
    {
        subdivisionDepth = 0;

        JFileChooser fc = new JFileChooser();
        javax.swing.filechooser.FileFilter ff = 
            new javax.swing.filechooser.FileFilter() {
            public boolean accept(File f)
            {
                if (f.isDirectory()) {
                    return true;
                }

                String fName = f.getName();
                int dot = fName.lastIndexOf(".");

                if (dot == -1) {
                    return false;
                }

                String suffix = fName.substring(dot);
                //System.out.println(suffix + " format detected");
                if (suffix.equals(".s3d") || suffix.equals(".obj") ||
                        suffix.equals(".m") || suffix.equals(".iv")) {
                    return true;
                }
                else {
                    return false;
                }
            }
            public String getDescription()
            {
                return "*.s3d, *.obj, *.m, *.iv";
            }
        };

        fc.setFileFilter(ff);
        fc.setAcceptAllFileFilterUsed(false);
        triangulate = new JCheckBox("Triangulate");
        triangulate.setSelected(true);
        fc.setAccessory(triangulate);

        int returnVal =
            fc.showOpenDialog(topLevel);

        if ( returnVal == JFileChooser.APPROVE_OPTION )
        {
            selectedFile = fc.getSelectedFile();
            String fName = selectedFile.getName();

            String suffix = fName.substring(fName.lastIndexOf("."));
            System.out.println(suffix + " format detected");
            try {
                if (suffix.equals(".s3d")) {
                    loadFromS3D(selectedFile);
                }
                else if (suffix.equals(".obj")) {
                    loadFromObj(selectedFile);
                }
                else if (suffix.equals(".m")) {
                    loadFromM(selectedFile);
                }
                else if (suffix.equals(".iv")) {
                    loadFromIV(selectedFile);
                }
                else {
                    //Unknown file type
                    selectedFile = null;
                    JOptionPane.showMessageDialog(topLevel, 
                            "Invalid file type");
                }
            }
            catch (Exception e) {
                    JOptionPane.showMessageDialog(topLevel, 
                            e.toString());
            }
        }
    }
    
    private void saveFile()
    {
        if (mesh == null) {
            JOptionPane.showMessageDialog(topLevel, "Must open a file first");
            return;
        }

        JFileChooser fc = new JFileChooser();
        javax.swing.filechooser.FileFilter ff = 
            new javax.swing.filechooser.FileFilter() {
            public boolean accept(File f)
            {
                if (f.isDirectory()) {
                    return true;
                }

                String fName = f.getName();
                int dot = fName.lastIndexOf(".");

                if (dot == -1) {
                    return false;
                }

                String suffix = fName.substring(dot);
                //System.out.println(suffix + " format detected");
                if (suffix.equals(".s3d") || suffix.equals(".iv")) {
                    return true;
                }
                else {
                    return false;
                }
            }
            public String getDescription()
            {
                return "*.s3d, *.iv";
            }
        };

        fc.setFileFilter(ff);
        fc.setAcceptAllFileFilterUsed(false);

        int returnVal =
            fc.showSaveDialog(topLevel);

        if ( returnVal == JFileChooser.APPROVE_OPTION )
        {
            File output = fc.getSelectedFile();
            String fName = output.getName();

            String suffix = fName.substring(fName.lastIndexOf("."));
            System.out.println(suffix + " format detected");
            try {
                if (suffix.equals(".s3d")) {
                    mesh.toS3D(output);
                }
                else if (suffix.equals(".iv")) {
                    mesh.toIV(output);
                }
                else {
                    //Unknown file type
                    JOptionPane.showMessageDialog(topLevel, 
                            "Invalid file type");
                }
            }
            catch (Exception e) {
                    JOptionPane.showMessageDialog(topLevel, 
                            e.toString());
            }
        }
    }

    private void selectAlgorithm()
    {
        JFileChooser fc = new JFileChooser();
        javax.swing.filechooser.FileFilter ff = 
            new javax.swing.filechooser.FileFilter() {
            public boolean accept(File f)
            {
                if (f.isDirectory()) {
                    return true;
                }

                String fName = f.getName();
                int dot = fName.lastIndexOf(".");

                if (dot == -1) {
                    return false;
                }

                String suffix = fName.substring(dot);
                if (suffix.equals(".class")) {
                    return true;
                }
                else {
                    return false;
                }
            }
            public String getDescription()
            {
                return "class file";
            }
        };

        fc.setFileFilter(ff);
        fc.setAcceptAllFileFilterUsed(false);

        int returnVal =
            fc.showOpenDialog(topLevel);

        if ( returnVal == JFileChooser.APPROVE_OPTION )
        {
            File classFile = fc.getSelectedFile();

            try {
                int dot = classFile.getName().indexOf(".class");
                String className = "";
                if (dot == -1) {
                    className = classFile.getName();
                }
                else {
                    className = 
                        classFile.getName().substring(0, dot);
                }

                algorithmClass = Class.forName(className);
                selectedAlgorithm = 
                    (SubdivisionAlgorithm)algorithmClass.newInstance();
            }
            catch (ClassCastException cce) {
                    JOptionPane.showMessageDialog(topLevel, 
                            "Invalid algorithm class");
            }
            catch (Exception e) {
                    JOptionPane.showMessageDialog(topLevel, 
                            e.toString());
            }
        }
    }

    private void updateStatus()
    {
        if (selectedFile != null) {
            fileOpenedTL.setText(selectedFile.toString());
        }
        if (algorithmClass != null) {
            algorithmSelectedTL.setText(algorithmClass.toString());
        }
        if (mesh != null) {
            numPolygonsTL.setText("" + mesh.numFaces());
        }

        subdivisionDepthTL.setText("" + subdivisionDepth);
    }

    private JMenuBar menuBarSetup()
    {
        JMenuBar menuBar = new JMenuBar();
        JMenu fileMenu = new JMenu("File");
        JMenu optionMenu = new JMenu("Tools");

        menuBar.add(fileMenu);
        menuBar.add(optionMenu);

        JMenuItem openItem = new JMenuItem();
        openItem.setAction(new AbstractAction()
                    {
                        public void actionPerformed(ActionEvent e)
                        {
                            openFile();
                            updateStatus();
                        }
                    });
        openItem.setText("Open");

        JMenuItem saveItem = new JMenuItem();
        saveItem.setAction(new AbstractAction()
                    {
                        public void actionPerformed(ActionEvent e)
                        {
                            saveFile();
                        }
                    });
        saveItem.setText("Save");

        JMenuItem quitItem = new JMenuItem();
        quitItem.setAction(new AbstractAction()
                    {
                        public void actionPerformed(ActionEvent e)
                        {
                            System.exit(0);
                        }
                    });
        quitItem.setText("Quit");

        JMenuItem algorithmItem = new JMenuItem();
        algorithmItem.setAction(new AbstractAction()
                    {
                        public void actionPerformed(ActionEvent e)
                        {
                            selectAlgorithm();
                            updateStatus();
                        }
                    });
        algorithmItem.setText("Algorithm");

        // add a "Open" item
        fileMenu.add(openItem);

        // add a "Save" item
        fileMenu.add(saveItem);

        // add a "Quit" item
        fileMenu.add(quitItem);

        //add a "Select Algorithm"  item
        optionMenu.add(algorithmItem);

        return menuBar;
    }

    private Container controlPanelSetup()
    {
        JPanel controlPanel = new JPanel();
        controlPanel.setLayout(new BorderLayout());

        JPanel labelPane = new JPanel();
        labelPane.setLayout(new GridLayout(0, 1));
        labelPane.add(new JLabel("File Selected :"));
        labelPane.add(new JLabel("Algorithm Selected :"));
        labelPane.add(new JLabel("Subdivision Depth :"));
        labelPane.add(new JLabel("Number of Polygons :"));

        JPanel textFieldPane = new JPanel();
        textFieldPane.setLayout(new GridLayout(0, 1));
        numPolygonsTL = new JTextField(30);
        numPolygonsTL.setEditable(false);
        subdivisionDepthTL = new JTextField(30);
        subdivisionDepthTL.setEditable(false);
        algorithmSelectedTL = new JTextField(30);
        algorithmSelectedTL.setEditable(false);
        fileOpenedTL = new JTextField(30);
        fileOpenedTL.setEditable(false);
        textFieldPane.add(fileOpenedTL);
        textFieldPane.add(algorithmSelectedTL);
        textFieldPane.add(subdivisionDepthTL);
        textFieldPane.add(numPolygonsTL);

        JPanel statusPanel = new JPanel();
        statusPanel.setBorder(BorderFactory.createEmptyBorder(20, 20,
                    20, 20));
        statusPanel.setLayout(new BorderLayout());
        statusPanel.add(labelPane, BorderLayout.CENTER);
        statusPanel.add(textFieldPane, BorderLayout.EAST);


        JPanel subdividePanel = new JPanel();
        subdividePanel.setLayout(new BorderLayout());
        JButton subdivideButton = new JButton();
        subdividePanel.add(subdivideButton, BorderLayout.SOUTH);
        subdivideButton.setAction(new AbstractAction()
                {
                    public void actionPerformed(ActionEvent e)
                    {
                        subdivide();
                        updateStatus();
                    }
                });
        subdivideButton.setText("Subdivide!");

        controlPanel.add(statusPanel, BorderLayout.WEST);
        controlPanel.add(subdividePanel, BorderLayout.EAST);

        return controlPanel;
    }

    public SubdivisionController()
    {
        topLevel = new JFrame();
        topLevel.setTitle(APP_TITLE);

        topLevel.addWindowListener( new WindowAdapter()
                {public void windowClosing(WindowEvent e) {System.exit(0);}} );

        topLevel.setJMenuBar(menuBarSetup());
        topLevel.getContentPane().add(controlPanelSetup());

        topLevel.pack();
        topLevel.setVisible(true);
    }

    public static void main(String[] args)
    {
        SubdivisionController subdivisionController = 
            new SubdivisionController();
    }
}
