import java.util.*;
import javax.vecmath.*;

public class Face
{
    Object attribute;
    HalfEdge edge;
    Vector3d normal = null;


    Face(Object attribute)
    {
        this.attribute = attribute;
    }

    HalfEdge getEdge ()
    {
        return edge;
    }

    public Vector3d getNormal()
    {
        if (this.normal == null) {
            return null;
        }
        else {
            return new Vector3d(this.normal);
        }
    }

    public void setNormal(Vector3d normal)
    {
        this.normal = new Vector3d(normal);
    }

    /*
    public Vector3d calcNormal()
    {
        if (normal == null) {
            Vector3d v1 = null;
            Vector3d v2 = null;
            Point3d p0 = null;
            Point3d p1 = null;
            Point3d p2 = null;
            this.normal = new Vector3d();

            HalfEdge he = edge;
            p1 = new Point3d((Point3d)vertexIndex.get(
                        he.getVertexHashCode()));
            HalfEdge prevHE = he.next;
            HalfEdge prevPrevHE = he.next;

            while (prevHE.next != he) {
                prevPrevHE = prevHE;
                prevHE = prevHE.next;
            }
            p2 = new Point3d((Point3d)vertexIndex.get(
                        prevPrevHE.getVertexHashCode()));
            p0 = new Point3d((Point3d)vertexIndex.get(
                        prevHE.getVertexHashCode()));
            v1 = new Vector(p1);
            v1.sub(p0);
            v2 = new Vector(p2);
            v2.sub(p0);
            v1.cross(v2);

            this.normal = v1;
        }

        return new Vector3d(normal);
    }
    */
    

    public void setAttribute(Object attribute)
    {
        this.attribute = attribute;
    }

    public Object getAttribute()
    {
        return attribute;
    }

    public boolean isBoundary()
    {
        Iterator it = halfEdgeIterator();
        while (it.hasNext()) {
            HalfEdge he = (HalfEdge)it.next();
            if (he.sym == null) {
                return true;
            }
        }
        return false;
    }

    /*
     * VertexIter: iterate over the face's vertices.
     * HalfedgeIter: iterate over the face's halfedges.
     * EdgeIter: iterate over the face's edges.
     * FaceIter: iterate over all edge-neighboring faces.
     */

    public Iterator vertexIterator()
    {
        return new Iterator()
        {
            HalfEdge startEdge = edge;
            HalfEdge curEdge = startEdge;

            boolean finished = false;

            public boolean hasNext()
            {
                return !finished;
            }

            /**
             * @return VertexHashCode
             */
            public Object next()
            {
                if (finished) {
                    throw new NoSuchElementException();
                }

                Object ret = 
                    new Integer(curEdge.getVertexHashCode());

                curEdge = curEdge.next;
                if (curEdge == startEdge) {
                    finished = true;
                }

                return ret;
            }

            public void remove()
            {
                throw new UnsupportedOperationException();
            }
        };
    }

    public Iterator halfEdgeIterator()
    {
        return new Iterator()
        {
            HalfEdge startEdge = edge;
            HalfEdge curEdge = startEdge;

            boolean finished = false;

            public boolean hasNext()
            {
                return !finished;
            }

            /**
             * @return HalfEdge
             */
            public Object next()
            {
                if (finished) {
                    throw new NoSuchElementException();
                }

                Object ret = curEdge;

                curEdge = curEdge.next;
                if (curEdge == startEdge) {
                    finished = true;
                }

                return ret;
            }

            public void remove()
            {
                throw new UnsupportedOperationException();
            }
        };
    }

    /**
     * @deprecated
     */
    public Iterator FaceIterator()
    {
        return new Iterator()
        {
            Iterator myHalfEdgeIterator = halfEdgeIterator();

            public boolean hasNext()
            {
                return myHalfEdgeIterator.hasNext();
            }

            /**
             * @return Face
             */
            public Object next()
            {
                HalfEdge curEdge = (HalfEdge)myHalfEdgeIterator.next();
                return curEdge.sym.face;
            }

            public void remove()
            {
                throw new UnsupportedOperationException();
            }
        };
    }
}
