import javax.vecmath.*;
import java.util.*;

public final class VertexHandle
{
    private Map vertexIndex;
    private Map vertexToEdgeMap;
    private Map vertexToAttributeMap;

    private int vertexHashCode;

    Vector3d normal = null;

    VertexHandle(Map vertexIndex, Map vertexToEdgeMap, Map vertexToAttributeMap,
            int vertexHashCode)
    {
        this.vertexIndex = vertexIndex;
        this.vertexToEdgeMap = vertexToEdgeMap;
        this.vertexHashCode = vertexHashCode;
        this.vertexToAttributeMap = vertexToAttributeMap;
    }

    public Vector3d calcNormal()
    {
        if (this.normal == null) {
            //double totalAngle = 0;
            //Map angleToNormal = new HashMap();
            List angleNormalList = new LinkedList();
            Iterator heIt = oHalfEdgeListIterator();
            Vector3d v1 = null;
            Vector3d v2 = null;
            Point3d p0 = null;
            Point3d p1 = null;
            Point3d p2 = null;
            this.normal = new Vector3d();
            class AngleNormalPair
            {
                double angle;
                Vector3d n;
            }

            //First retrieve the angle
            while (heIt.hasNext()) {
                HalfEdge he = (HalfEdge)heIt.next();

                /*
                System.out.println("vertex in question is " + getVertex());
                System.out.println("p0 = " + p0);
                System.out.println("p1 = " + p1);
                System.out.println("p2 = " + p2);
                System.out.println("v1 = " + v1);
                System.out.println("v2 = " + v2);
                */

                //double angle = v1.angle(v2);
                Vector3d faceNormal = he.face.getNormal();
                if (faceNormal == null) {
                    HalfEdge prevHE = he.next;
                    HalfEdge prevPrevHE = he.next;

                    while (prevHE.next != he) {
                        prevPrevHE = prevHE;
                        prevHE = prevHE.next;
                    }

                    p1 = new Point3d((Point3d)vertexIndex.get(
                                new Integer(he.getVertexHashCode())));
                    p2 = new Point3d((Point3d)vertexIndex.get(
                                new Integer(prevPrevHE.getVertexHashCode())));
                    p0 = new Point3d((Point3d)vertexIndex.get(
                                new Integer(prevHE.getVertexHashCode())));
                    v1 = new Vector3d(p1);
                    v1.sub(p0);
                    v2 = new Vector3d(p2);
                    v2.sub(p0);

                    v1.cross(v1, v2);
                    //v1.normalize();
                    he.face.setNormal(v1);
                    faceNormal = v1;
                }
                //System.out.println("face normal is = " + faceNormal);
                //angleToNormal.put(new Double(angle), faceNormal);
                AngleNormalPair anp = new AngleNormalPair();
                //anp.angle = angle;
                anp.n = faceNormal;
                angleNormalList.add(anp);

                //totalAngle += angle;
                //System.out.println("angle is " + angle);
            }//for all out-going half edges from this vertex
            //System.out.println("total angle is " + totalAngle);

            //now sum up all face vertices and weight them according to the
            //adjacent angle
            //Set angles = angleToNormal.keySet();
            Iterator anglesIt = angleNormalList.iterator();
            
            while (anglesIt.hasNext()) {
                AngleNormalPair anp = (AngleNormalPair)anglesIt.next();
                //Double angle = (Double)anglesIt.next();
                //System.out.println("angle is " + anp.angle);
                //double weight = anp.angle/totalAngle;
                //double weight = 1.0/(double)angleNormalList.size();
                //System.out.println("weight is " + weight);
                Vector3d faceNormal = new Vector3d(anp.n);
                //faceNormal.scale(weight);
                this.normal.add(faceNormal);
                /*
                double weight = angle.doubleValue()/totalAngle;
                Vector3d faceNormal = (Vector3d)angleToNormal.get(angle);
                faceNormal.scale(weight);
                this.normal.add(faceNormal);
                */
            }

            this.normal.normalize();
            /*
            System.out.println("total angle is " + totalAngle);
            System.out.println("-------------------------------------");
            */
            //System.out.println("-------------------------------------");
        }//if we didn't calculate the normal of this vertex before

        return new Vector3d(this.normal);
    }

    public void setAttribute(Object attribute) 
    {
        vertexToAttributeMap.put(new Integer(this.vertexHashCode), attribute);
    }

    public Object getAttribute()
    {
        return vertexToAttributeMap.get(new Integer(this.vertexHashCode));
    }

    public int getValence()
    {
        Iterator vIt = vertexIterator();
        int valence = 0;
        //FIXME : endless loop
        while (vIt.hasNext()) {
            vIt.next();
            valence ++;
        }
        return valence;
    }

    public Point3d getVertex()
    {
        return (Point3d)vertexIndex.get(new Integer(vertexHashCode));
    }

    /*
    public void setVertex(Point3d newVertex)
    {
        vertexIndex.put(vertexHashCode, newVertex);
    }
    */

    public int getVertexHashCode()
    {
        return vertexHashCode;
    }

    /**
     * gets the first out-going half edge on the out-going half edge list
     */
    public HalfEdge getEdge()
    {
        List el = getEdgeList();
        if (el == null) {
            return null;
        }
        else {
            return (HalfEdge)el.get(0);
        }
    }

    public List getEdgeList()
    {
        List ret = 
            (List)vertexToEdgeMap.get(new Integer(vertexHashCode));

        if (ret == null) {
            return null;
        }
        else {
            return ret;
        }
    }

    void setEdgeList(List edgeList)
    {
        vertexToEdgeMap.put(new Integer(vertexHashCode), edgeList);
    }

    /**
     * A vertex is a bounday vertex if its set of incident edges doesn't form
     * a cycle.
     */
    boolean isBoundary()
    {
        Iterator it = oHalfEdgeIterator();
        while (it.hasNext()) {
            if (((HalfEdge)it.next()).sym == null) {
                return true;
            }
        }

        return false;
    }

    /*
     * vertexIterator: iterate over all neighboring vertices.
     * iHalfedgeIterator: iterate over all incoming halfedges.
     * oHalfedgeIterator: iterate over all outgoing halfedges.
     * faceIterator: iterate over all adjacent faces.
     */

    /**
     * Will NOT handle "bow-ties" properly
     */
    public Iterator vertexIterator()
    {
        return vertexIterator(null);
    }

    public Iterator vertexIterator(HalfEdge startingHE)
    {
        final HalfEdge startHE = startingHE;

        return new Iterator()
        {
            Iterator myHEIt;
            Object onDeck;
            boolean finished;

            {
                myHEIt = oHalfEdgeIterator(startHE);
                finished = false;
                onDeck = null;
            }

            public boolean hasNext()
            {
                if (onDeck != null) {
                    return true;
                }
                else {
                    return myHEIt.hasNext();
                }
            }

            /**
             * @return VertexHashCode
             */
            public Object next()
            {
                HalfEdge he;
                Object ret;
                
                if (onDeck == null) {
                    he = (HalfEdge)myHEIt.next();
                    ret = new Integer(he.getVertexHashCode());
                    if (he.next.next.sym == null) {
                        //we hit a boundary
                        onDeck = new Integer(he.next.getVertexHashCode());
                    }
                }
                else {
                    ret = onDeck;
                    onDeck = null;
                }


                return ret;
            }

            public void remove()
            {
                throw new UnsupportedOperationException();
            }
        };
    }

    /**
     * Iterators over all half edges comming into this vertex
     */
    public Iterator iHalfEdgeIterator()
    {
        return new Iterator()
        {
            Iterator myOHalfEdgeIterator =
                oHalfEdgeIterator();

            public boolean hasNext()
            {
                return myOHalfEdgeIterator.hasNext();
            }

            /**
             * @return HalfEdge
             */
            public Object next()
            {
                HalfEdge oHE =
                    (HalfEdge)myOHalfEdgeIterator.next();
                return oHE.next.next;
            }

            public void remove()
            {
                throw new UnsupportedOperationException();
            }
        };
    }

    /**
     * Iterators over all half edges comming out of this vertex
     */
    public Iterator oHalfEdgeListIterator()
    {
        return new Iterator()
        {
            Iterator listIt;

            {
                //Instance initializer
                List edgeList = getEdgeList();
                if (edgeList == null) {
                    listIt = null;
                }
                else {
                    listIt = edgeList.iterator();
                }
            }
            public boolean hasNext()
            {
                if (listIt == null) {
                    return false;
                }
                return listIt.hasNext();
            }

            /**
             * @return HalfEdge
             */
            public Object next()
            {
                return listIt.next();
            }

            public void remove()
            {
                throw new UnsupportedOperationException();
            }
        };
    }

    /**
     * Iterators over all half edges comming out of this vertex
     */
    public Iterator oHalfEdgeIterator()
    {
        return oHalfEdgeIterator(null);
    }

    public Iterator oHalfEdgeIterator(HalfEdge startingHE)
    {
        final HalfEdge startHE = startingHE;

        return new Iterator()
        {
            HalfEdge startEdge = getEdge();
            HalfEdge curEdge = startEdge;
            boolean finished = false;

            {
                if (startHE != null) {
                    List el = getEdgeList();
                    Iterator elIt = el.iterator();
                    while (elIt.hasNext()) {
                        startEdge = (HalfEdge)elIt.next();
                        if (startEdge == startHE) {
                            break;
                        }
                    }
                }
                else {
                    startEdge = getEdge();
                }

                curEdge = startEdge;
                finished = false;


                //Sanity check
                /*
                HalfEdge check = curEdge;
                int a;
                for (a = 0; a < 50; a++ ) {
                    if (check == null) {
                        System.out.println("WTF?????????????????");
                    }
                    HalfEdge startCheck = check;
                    while (!check.vertexHashCode.equals(vertexHashCode)) {
                        check = check.next;
                    }
                    if (check.next != startCheck) {
                        System.out.println("next pointer is fucked up");
                    }
                    check = check.sym;
                    if (check == startEdge) {
                        //System.out.println("alright " + a);
                        break;
                    }
                }
                if (a >= 40) {
                    System.out.println("Shit!!!!!!!!!!!!!!!!!!!!!!!!!!!!" + a);
                }
                */
            }

            public boolean hasNext()
            {
                return !finished;
            }

            /**
             * @return HalfEdge
             */
            public Object next()
            {
                if (finished) {
                    throw new NoSuchElementException();
                }
                
                Object ret = curEdge;

                while (curEdge.vertexHashCode != vertexHashCode) {
                    curEdge = curEdge.next;
                }

                if (curEdge.sym == null) {
                    //System.out.println("BOUNDARY?!?!?!?!?!?!?");
                    while (curEdge.next.sym != null) {
                        //previous incomming edge
                        curEdge = curEdge.next.sym;
                    }
                    //the boundary out-going edge
                    curEdge = curEdge.next;
                }
                else {
                    curEdge = curEdge.sym;
                }

                if (curEdge == startEdge) {
                    finished = true;
                }

                return ret;
            }

            public void remove()
            {
                throw new UnsupportedOperationException();
            }
        };
    }
}
